import json
import logging

from johnnydep.lib import JohnnyDist, flatten_deps

from verinfast.dependencies.walkers.classes import Walker, Entry


def parseFile(filename="requirements.txt", ret=False):
    dists = []
    default_fields = [
        "name",
        "summary",
        "specifier",
        "requires",
        "required_by",
        "license"
    ]
    with open(filename) as file:
        for line in file:
            if line.find("#") >= 0:
                line = line[0:line.find("#")]
            stripped_line = line.rstrip()
            if stripped_line[0:2] == '--' or not stripped_line:
                pass
            else:
                try:
                    dists.append(JohnnyDist(
                            stripped_line,
                            ignore_errors=True,
                        )
                    )
                except Exception as error:
                    # handle the exception, hiding for now
                    logger = logging.getLogger()
                    logger.disabled = True
                    logger.exception(error)
                    logger.disabled = False
                    pass

    data = []
    for idx, d in enumerate(dists):
        deps = flatten_deps(d)
        data += deps

    # This is the worst line of code I've ever written.
    output = [
        d for dep in data
        for d in dep.serialise(fields=default_fields, recurse=False)
    ]

    dup_check = {}
    for idx, o in enumerate(output):
        k = o["name"]+o["specifier"]
        if k in dup_check:
            output[dup_check[k]]["required_by"].append(k)
            output.remove(o)
        else:
            dup_check[k] = idx

    result = json.dumps(output, indent=2, default=str, separators=(",", ": "))
    if not ret:
        print(result)
    else:
        return output


class PyWalker(Walker):
    def parse(self, file: str, expand=False, ret=True):
        temp = parseFile(filename=file, ret=ret)
        for el in temp:
            el["source"] = "pip"
        self.entries = [Entry(**entry) for entry in temp]



import json
from unittest.mock import patch
import os
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog
import verinfast.user


@patch('verinfast.user.__get_input__', return_value='y')
def test_gcp_scan(self):
    assert verinfast.user.initial_prompt is not None
    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    results_dir = test_folder.joinpath("results").absolute()
    cfg_path = test_folder.joinpath("gcp_conf.yaml").absolute()
    agent = Agent()
    config = Config(cfg_path=str(cfg_path))
    assert config.cfg_path == str(cfg_path)
    assert config.config is not FileNotFoundError
    assert "modules" in config.config
    sub_id = config.config["modules"]["cloud"][0]["account"]
    config.output_dir = str(results_dir)
    config.runGit = False
    agent.config = config
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    assert agent.config.output_dir == str(results_dir)
    assert agent.config.dry is False
    assert agent.config.runGit is False
    assert len(agent.config.modules.cloud) > 0
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
    assert Path(results_dir).exists() is False
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.scan()
    assert Path(results_dir).exists() is True
    with open(results_dir.joinpath(f"gcp-instances-{sub_id}.json")) as f:
        instances = json.load(f)
        assert len(instances["data"]) >= 1
    with open(
        results_dir.joinpath(f"gcp-instances-{sub_id}-utilization.json")
    ) as f:
        utilization = json.load(f)
        for u in utilization["data"]:
            assert isinstance(u["id"], str)



from datetime import date
import inspect
import os
import re
from glob import glob
import subprocess
import time

from typing import List

STD_EXCLUDE_LIST = [
                "**/.git/**",
                "**/node_modules/**",
                "build/**",
                "dist/**",
                "env/**",
                "venv/**",
                "**/__pycache__/**",
                "**/*.pyc",
                "**/*.pyo",
                "**/*.log",
                "**/*.swp",
                "**/*.swo",
                "**/*.DS_Store",
                "**/.idea/**"
            ]


"""
newline = "\n" #
TODO - Set to system appropriate newline character.
This doesn't work with modernmetric
"""


def list_files(
        path: str,
        incl: List[glob] = ["*.*"],
        excl: List[glob] = STD_EXCLUDE_LIST
        ):
    files = []
    excluded = []
    for i in incl:
        files += glob(os.path.join(path, i))
    for e in excl:
        excluded += glob(os.path.join(path, e))
    files = [f for f in files if f not in excluded]
    return files


def std_exec(cmd: List[str], log=None):
    try:
        return subprocess.check_output(cmd).decode('utf-8')
    except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
        try:
            return subprocess.check_output(
                cmd, shell=True).decode(encoding='latin-1')
        except Exception as e2:
            if log is not None:
                log(tag="std_exec replace Error", msg=f"{e2}, {cmd}")
            else:
                print(f"std_exec replace Error: {e2}, {cmd}")
            if log is not None:
                log(tag="std_exec Error", msg=f"{e}, {cmd}")
            else:
                print(f"std_exec Error: {e}, {cmd}")
            return ''


def escapeChars(text: str):
    fixedText = re.sub(r'([\"\{\}])', r'\\\1', text)
    return fixedText


def trimLineBreaks(text: str):
    return text.replace("\n", "").replace("\r", "")


# Truncate large strings for display
def truncate(text, length=100):
    testStr = str(text)  # Supports passing in Lists and other types
    return ((testStr[:length] + '..') if len(testStr) > length else testStr)


def truncate_children(
            obj: dict | list,
            log,
            excludes=[],
            max_length=30,
            recursion_depth=0
        ):
    if isinstance(obj, dict):
        for k in obj:
            v = obj[k]
            if k in excludes:
                pass
            elif isinstance(v, str):
                obj[k] = v[0:max_length]
            elif (
                isinstance(v, int) or
                isinstance(v, float) or
                isinstance(v, bool)
            ):
                pass
            else:
                obj[k] = truncate_children(
                    obj[k],
                    log,
                    excludes=excludes,
                    max_length=max_length,
                    recursion_depth=recursion_depth + 1
                )
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            if isinstance(v, str):
                obj[i] = v[0:max_length]
            elif (
                isinstance(v, int) or
                isinstance(v, float) or
                isinstance(v, bool)
            ):
                pass
            else:
                obj[i] = truncate_children(
                    obj[i],
                    log,
                    excludes=excludes,
                    max_length=max_length,
                    recursion_depth=recursion_depth + 1
                )

    return obj


# Chunk a list
def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class DebugLog:
    def __init__(self, path: str, debug: bool = False):
        self.path = path
        self.logFile = os.path.join(path, "log.txt")
        self.debug = debug

    def log(self, msg, tag=None, display=False, timestamp=True):
        if tag is None and timestamp:
            s = inspect.stack()[1]

            tag = f"{s.filename}@{s.lineno} {s.function}"
        if timestamp:
            d = date.today()
            tag = f"{d} {time.strftime('%H:%M:%S', time.localtime())} {tag}"
        if tag != "":
            output = f"{tag}: {msg}"
        else:
            output = f"{msg}"

        with open(self.logFile, 'a') as f:
            f.write(output+"\n")
        if display or self.debug:
            print(output)


# Returns a dict of the repo name, repo url and branch from the original url
def get_repo_name_url_and_branch(repo_url: str):
    # The regex pattern captures the entire URL
    # and the repository name after the last '/'

    repo_url = repo_url.strip()
    branch = None

    split_url = repo_url.split("@")

    # HTTPS URLs
    if repo_url.startswith("https://") or repo_url.startswith("http://"):
        # Check for branch in URL
        if len(split_url) == 2:
            branch = split_url[1]
            repo_url = split_url[0]
        elif len(split_url) == 3:
            # Username and password in URL
            # e.g. https://user:password@my.githost.com/repo@branch
            branch = split_url[2]
            # Remove branch so that the repo name can be extracted
            # without colliding with /'s in branch name
            repo_url = repo_url.replace(f"@{branch}", "")
    else:
        # SSH URLs
        # Check for branch in URL
        if len(split_url) > 2:
            branch = split_url[2]
            # Remove branch so that the repo name can be extracted
            # without colliding with /'s in branch name
            repo_url = repo_url.replace(f"@{branch}", "")
            print(f"Repo URL after removing branch: {repo_url}")
            print(f"@{branch}")

    # Match repo_name to after the last '/' in the cleaned URL
    repo_name = repo_url.rsplit('/', 1)[-1]

    return {
        "repo_name": repo_name,
        "repo_url": repo_url,
        "branch": branch
    }



from verinfast.dependencies.walkers.classes import Walker, Entry
import xml.etree.ElementTree as ET


class MavenWalker(Walker):
    # pkg:maven/springframework/spring@1.2.6
    # mvn license:aggregate-add-third-party
    # mvn dependency:analyze
    # mvn dependency-check:check
    # mvn org.sonatype.ossindex.maven:ossindex-maven-plugin:audit -f pom.xml
    # <project>
    # ...
    # <properties>
    # <mavenVersion>3.0</mavenVersion>
    # </properties>

    # <dependencies>
    # <dependency>
    #     <groupId>org.apache.maven</groupId>
    #     <artifactId>maven-artifact</artifactId>
    #     <version>${mavenVersion}</version>
    # </dependency>
    # <dependency>
    #     <groupId>org.apache.maven</groupId>
    #     <artifactId>maven-core</artifactId>
    #     <version>${mavenVersion}</version>
    # </dependency>
    # </dependencies>
    # ...
    # </project>

    def parse(self, file: str, expand=False):
        tree = ET.parse(file)
        # root = tree.getroot()
        dependencies = tree.findall('dependencies/dependency')
        for d in dependencies:
            groupId = d.find("groupId")
            artifcact = d.find("artifactId")
            version = d.find("version")
            g = groupId.text
            a = artifcact.text
            v = version.text
            n = g + "/" + a
            e = Entry(
                name=n,
                source='maven',
                specifier="=="+v
            )
            self.entries.append(e)

    def expand(self, file):
        raise Exception("No expansion for this Walker")



from verinfast.dependencies.walkers.classes import Walker, Entry


class DockerWalker(Walker):
    def clean_str(self, s: str | bytes):
        bad_chars = ["\n", "\r", "\t"]
        s = str(s).lower().strip()
        for char in bad_chars:
            s = s.replace(char, "")
        return s

    def parse_address(
            self,
            address: str,
            file: str,
            source: str = "Dockerfile"
            ) -> Entry:
        name = address
        specifier = "*"

        if "/" in name:
            last_slash = name.rfind("/")
            sub_str = name[last_slash+1:]
            if "@" in name:
                [name, specifier] = name.split("@")
            elif ":" in sub_str:
                specifier = sub_str.split(":")[1]
                name = name[0:last_slash]

        elif "@" in name:
            [name, specifier] = name.split("@")

        elif ":" in name:
            [name, specifier] = name.split(":")

        e = Entry(
            name=name,
            specifier=specifier,
            source=source,
            required_by=file
        )
        return e

    def parse(self, file: str, expand=False):
        with open(file, 'r') as f:
            string = f'\nDockerfile found {file}:\n\n\n {str(f.read())} \n\n\n'
            self.log(timestamp=False, tag=None, msg=string)
        with open(file, 'r') as f:
            lines = f.readlines()
            for line in lines:
                line = self.clean_str(line)
                if line.startswith('from '):
                    name = line.split(' ')[1]
                    e = self.parse_address(name, file)
                    self.entries.append(e)
                elif line.startswith("image:"):
                    name = line[6:]
                    e = self.parse_address(name, file, source="docker-compose")
                    self.entries.append(e)

    def expand(self, file):
        raise Exception("No expansion for this Walker")



import json
from unittest.mock import patch
import os
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog
import verinfast.user


@patch('verinfast.user.__get_input__', return_value='y')
def test_aws_scan(self):
    sub_id = 436708548746
    assert verinfast.user.initial_prompt is not None
    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    results_dir = test_folder.joinpath("results").absolute()
    cfg_path = test_folder.joinpath("aws_conf.yaml").absolute()
    agent = Agent()
    config = Config(cfg_path=str(cfg_path))
    assert config.cfg_path == str(cfg_path)
    assert config.config is not FileNotFoundError
    assert "modules" in config.config
    config.output_dir = str(results_dir)
    config.runGit = False
    agent.config = config
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    assert agent.config.output_dir == str(results_dir)
    assert agent.config.dry is False
    assert agent.config.runGit is False
    assert len(agent.config.modules.cloud) > 0
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
    assert Path(results_dir).exists() is False
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.scan()
    assert Path(results_dir).exists() is True
    with open(results_dir.joinpath(f"aws-cost-{sub_id}.json")) as f:
        costs = json.load(f)
        assert len(costs["data"]) >= 100
    # Make sure "aws-cost-foo.json" doesn't exist
    bad_costs_file = Path(results_dir.joinpath("aws-cost-foo.json"))
    assert not bad_costs_file.is_file()
    instancesFile = results_dir.joinpath(f"aws-instances-{sub_id}.json")
    print("instancesFile", instancesFile)
    with open(instancesFile) as f:
        instances = json.load(f)
        assert len(instances["data"]) >= 5
    # Make sure "aws-instances-foo.json" doesn't exist
    bad_instances_file = Path(results_dir.joinpath("aws-instances-foo.json"))
    assert not bad_instances_file.is_file()
    with open(
        results_dir.joinpath(f"aws-instances-{sub_id}-utilization.json")
    ) as f:
        utilization = json.load(f)
        v = []
        for u in utilization["data"]:
            if u["id"] == "i-0a3493d5abcdfba4b":
                v = u["metrics"]
        assert len(v) >= 450
    # Make sure "aws-instances-foo-utilization.json" doesn't exist
    bad_utilization_file = Path(
        results_dir.joinpath("aws-instances-foo-utilization.json")
    )
    assert not bad_utilization_file.is_file()
    storageFile = results_dir.joinpath(f"aws-storage-{sub_id}.json")
    print("storageFile", storageFile)
    try:
        with open(storageFile) as f:
            storage = json.load(f)
            v = 0
            for u in storage["data"]:
                if u["name"] == "startupos-test-bucket":
                    v = u["size"]
            assert v >= 262183
    except Exception as e:
        print(e)
        print("Missing the storage file")
        print(os.listdir(results_dir))
    # Make sure "aws-storage-foo.json" doesn't exist
    bad_storage_file = Path(results_dir.joinpath("aws-storage-foo.json"))
    assert not bad_storage_file.is_file()


@patch('verinfast.user.__get_input__', return_value='y')
def test_aws_dash(self):
    sub_id = "436708548746"
    assert verinfast.user.initial_prompt is not None
    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    results_dir = test_folder.joinpath("results").absolute()
    cfg_path = test_folder.joinpath("aws_dash.yaml").absolute()
    agent = Agent()
    config = Config(cfg_path=str(cfg_path))
    assert config.cfg_path == str(cfg_path)
    assert config.config is not FileNotFoundError
    assert "modules" in config.config
    config.output_dir = str(results_dir)
    config.runGit = False
    agent.config = config
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    assert agent.config.output_dir == str(results_dir)
    assert agent.config.dry is False
    assert agent.config.runGit is False
    assert len(agent.config.modules.cloud) > 0
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
    assert Path(results_dir).exists() is False
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.scan()
    assert Path(results_dir).exists() is True
    with open(results_dir.joinpath(f"aws-cost-{sub_id}.json")) as f:
        costs = json.load(f)
        assert len(costs["data"]) >= 100
    with open(results_dir.joinpath(f"aws-instances-{sub_id}.json")) as f:
        instances = json.load(f)
        assert len(instances["data"]) >= 5
    with open(
        results_dir.joinpath(f"aws-instances-{sub_id}-utilization.json")
    ) as f:
        utilization = json.load(f)
        v = []
        for u in utilization["data"]:
            if u["id"] == "i-0a3493d5abcdfba4b":
                v = u["metrics"]
        assert len(v) >= 450
    with open(results_dir.joinpath(f"aws-storage-{sub_id}.json")) as f:
        storage = json.load(f)
        v = 0
        for u in storage["data"]:
            if u["name"] == "startupos-test-bucket":
                v = u["size"]
        assert v >= 262183



import json
from unittest.mock import patch
import os
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog
import verinfast.user
from verinfast.dependencies.walk import walk
from verinfast.dependencies.walkers.classes import Entry

file_path = Path(__file__)
test_folder = file_path.parent
docker_folder = test_folder.joinpath('fixtures/docker')
results_dir = test_folder.joinpath("results").absolute()
composer_config_path = str(test_folder.joinpath('composer.yaml').absolute())


def enabled_logger(flag=True):
    def silent(msg, **kwargs):
        return None

    def logger(msg, **kwargs):
        print(msg)
        print(kwargs)

    if flag:
        return logger
    else:
        return silent


def test_dockerfile_exists():
    docker_file_path = docker_folder.joinpath('Dockerfile')
    assert Path.exists(docker_file_path)


def test_walk():
    folder_path = test_folder.joinpath("fixtures/npm_walker").absolute()
    file_path = folder_path.joinpath('package.json')

    output_file = test_folder.joinpath("dependencies1.json")
    assert file_path.exists(), "Manifest doesn't exist"
    assert not output_file.exists(), "Results file exists"

    output_path = walk(
        path=folder_path,
        output_file=output_file,
        logger=enabled_logger(False)
        )

    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1
        first_dep = output[0]
        assert first_dep['name'] == 'simple-test-package'

    os.remove(output_path)
    assert not Path(output_path).exists()
    node_modules = folder_path.joinpath('node_modules')
    shutil.rmtree(node_modules)
    os.remove(folder_path.joinpath("package-lock.json"))
    return None


def test_entity():
    folder_path = test_folder.joinpath("fixtures/nuget_walker").absolute()
    file_path = folder_path.joinpath('test.csproj')
    output_file = test_folder.joinpath("dependencies2.json")
    assert file_path.exists(), "Manifest doesn't exist"
    shutil.rmtree(output_file, ignore_errors=True)
    try:
        os.remove(output_file)
    except:  # noqa
        pass
    assert not output_file.exists(), "Results file exists"

    output_path = walk(
        path=str(folder_path),
        output_file=output_file,
        logger=enabled_logger(False)
    )
    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1
        first_dep = output[0]
        e = Entry(**first_dep)
        assert e.license == "MIT"
        found_Cosmos = False
        for d in output:
            if d["name"] == "Microsoft.Azure.Cosmos":
                found_Cosmos = True
                assert d["license"] == "https://aka.ms/netcoregaeula"
        assert found_Cosmos
    os.remove(output_path)
    assert not Path(output_path).exists()

    return None


# def test_ruby():
#     output_path = walk(
#         path=test_folder.joinpath("fixtures/gemwalker"),
#         output_file="./dependencies.json",
#         logger=logger
#     )
#     with open(output_path) as output_file:
#         output = json.load(output_file)
#         assert len(output) >= 1
#         first_dep = output[0]
#         e = Entry(**first_dep)
#         assert e.license == "ISC"
#         found_rubocop = False
#         for d in output:
#             if d["name"] == "rubocop-ast":
#                 found_rubocop = True
#                 assert d["specifier"] == "*"
#         assert found_rubocop
#         found_aasm = False
#         for d in output:
#             if d["name"] == "aasm":
#                 found_aasm = True
#                 assert d["specifier"] == "*"
#         assert found_aasm
#         found_bad_source_type = False
#         found_source_single_quote = False
#         for d in output:
#             if type(d["source"]) is dict:
#                 found_bad_source_type = True
#             elif d["source"].find("'") != -1:
#                 found_source_single_quote = True
#         assert not found_bad_source_type
#         assert not found_source_single_quote

#     return None


# def test_python():
#     output_path = walk(
#         path=test_folder.joinpath("fixtures/python_walker"),
#         output_file="./dependencies.json",
#         logger=logger
#     )
#     with open(output_path) as output_file:
#         output = json.load(output_file)
#         assert len(output) >= 1
#         first_dep = output[0]
#         e = Entry(**first_dep)
#         assert e.license == "ISC"
#         found_azure_identity = False
#         for d in output:
#             if d["name"] == "azure-identity":
#                 found_azure_identity = True
#                 assert d["source"] == "pip"
#         assert found_azure_identity
#         found_azure_core = False
#         for d in output:
#             if d["name"] == "azure-core":
#                 found_azure_core = True
#                 assert d["source"] == "pip"
#         assert found_azure_core

#     return None


def test_docker():
    output_path = walk(
        path=docker_folder,
        output_file="./dependencies.json",
        logger=enabled_logger(False)
    )
    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1
        found_ubuntu_base_image = False
        for d in output:
            if d["name"] == "ubuntu":
                found_ubuntu_base_image = True
                assert d["specifier"] == "trusty"
                assert d["source"] == "Dockerfile"
        assert found_ubuntu_base_image
    os.remove(output_path)

    return None


def test_composer():
    folder_path = test_folder.joinpath("fixtures/composer_walker").absolute()
    file_path = folder_path.joinpath('composer.json')
    assert file_path.exists(), "Manifest doesn't exist"

    output_file_path = test_folder.joinpath("dependencies3.json")
    try:
        os.remove(output_file_path)
    except:  # noqa
        pass

    output_path = walk(
        path=folder_path,
        output_file=output_file_path,
        logger=enabled_logger(False)
    )
    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1

    os.remove(output_path)

    return None


def test_go():
    folder_path = test_folder.joinpath("fixtures/go_walker").absolute()
    file_path = folder_path.joinpath('go.sum')
    assert file_path.exists(), "Manifest doesn't exist"

    output_file_path = test_folder.joinpath("dependencies4.json")
    try:
        os.remove(output_file_path)
    except FileNotFoundError:  # noqa
        pass

    output_path = walk(
        path=folder_path,
        output_file=output_file_path,
        logger=enabled_logger(False)
    )
    with open(output_path) as output_file:
        output = json.load(output_file)
        assert len(output) >= 1
        assert output[0]['name'] == "cloud.google.com/go"
        assert output[0]['source'] == "Go"
        assert output[0]['specifier'] == "v0.26.0"

    os.remove(output_path)

    return None


@patch('verinfast.user.__get_input__', return_value='y')
def test_composer_config(self):
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    assert verinfast.user.initial_prompt is not None
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(composer_config_path)
    config.output_dir = results_dir
    agent.config = config
    agent.config.shouldUpload = False
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    agent.scan()
    assert Path(results_dir).exists() is True
    expected_files = [
        "CssToInlineStyles.git.dependencies.json",
        "laravel.git.dependencies.json"
    ]
    for f in expected_files:
        expected_file = results_dir.joinpath(f)
        assert expected_file.exists(), f"{expected_file} does not exist"



from verinfast.dependencies.walkers.classes import Walker, Entry


class GoWalker(Walker):
    def parse(self, file: str, expand=False):
        with open(file, "r") as file:
            for line in file:
                name, version, hash_string = line.split(" ")
                if version.endswith("go.mod"):
                    version = version.split("/")[0]
                e = Entry(
                    name=name,
                    specifier=version,
                    source="Go"
                )
                self.entries.append(e)



#!/usr/bin/env python3
from datetime import date
import json
import os
from pathlib import Path
import platform
import re
import shutil
import subprocess
import traceback
from uuid import uuid4

from modernmetric.__main__ import main as modernmetric

import httpx
from jinja2 import Environment, FileSystemLoader
from pygments_tsx.tsx import patch_pygments

from verinfast.utils.utils import DebugLog, std_exec, trimLineBreaks, escapeChars, truncate, truncate_children, get_repo_name_url_and_branch
from verinfast.upload import Uploader

from verinfast.cloud.aws.costs import runAws
from verinfast.cloud.aws.get_profile import find_profile
from verinfast.cloud.azure.costs import runAzure
from verinfast.cloud.aws.instances import get_instances as get_aws_instances
from verinfast.cloud.azure.instances import get_instances as get_az_instances
from verinfast.cloud.gcp.instances import get_instances as get_gcp_instances
from verinfast.cloud.aws.blocks import getBlocks as get_aws_blocks
from verinfast.cloud.azure.blocks import getBlocks as get_az_blocks
from verinfast.cloud.gcp.blocks import getBlocks as get_gcp_blocks
from verinfast.config import Config
from verinfast.user import initial_prompt, save_path, repeat_boolean_prompt

from verinfast.dependencies.walk import walk as dependency_walk

# from verinfast.pygments_patch import patch_pygments

patch_pygments()

today = date.today()

requestx = httpx.Client(http2=True, timeout=None)
uname = platform.uname()
system = uname.system
node = uname.node
release = uname.release
version = uname.version
machine = uname.machine


template_definition = {}

file_path = Path(__file__)
parent_folder = file_path.parent.absolute()
templates_folder = str(parent_folder.joinpath("templates"))
# str_path = str(parent_folder.joinpath('str_conf.yaml').absolute())

curr_dir = os.getcwd()
temp_dir = Path(os.path.expanduser('~/.verinfast/')).joinpath('temp_repo')


class Agent:
    def __init__(self):
        self.config = Config()
        os.makedirs(self.config.output_dir, exist_ok=True)
        self.debug = DebugLog(path=self.config.output_dir, debug=False)
        self.log = self.debug.log
        self.log(msg='', tag="Started")
        self.uploader = Uploader(self.config.upload_conf)
        self.up = self.uploader.make_upload_path
        self.config.upload_logs = initial_prompt()
        self.directory = save_path()

    # Takes a string and shows it to the user.
    # Also ensures that string is written to the logs.
    # No other decoration is allowed.
    def print_and_log(self, msg: str):
        self.log(msg=msg, tag="", display=True, timestamp=False)

    def create_template(self):
        if not self.config.dry:
            with open(f"{self.config.output_dir}/results.html", "w") as f:
                jinja_env = Environment(loader=FileSystemLoader(templates_folder))
                jinja_env.globals.update(zip=zip, sorted=sorted)
                output = jinja_env.get_template("results.j2").render(template_definition)
                f.write(output)

    def scan(self):
        if self.config.modules is not None:
            if self.config.modules.code is not None:

                if self.config.shouldUpload:
                    headers = {
                        'content-type': 'application/json',
                        'Accept-Charset': 'UTF-8',
                    }
                    get_url = self.uploader.make_upload_path("scan_id", report=self.config.reportId)
                    self.log(msg=f"{self.config.baseUrl}{get_url}", tag="Report Run Id Fetch", display=True)
                    response = requestx.get(f"{self.config.baseUrl}{get_url}", headers=headers)
                    self.scanId = response.text.replace("'", "").replace('"', '')
                    if self.scanId and self.scanId != '':
                        self.log(msg=self.scanId, tag="Report Run Id", display=True)
                    else:
                        raise Exception(f"{self.scanId} returned for failed report Id fetch.")
                else:
                    print("ID only fetched for upload")
                self.scanRepos()
            if self.config.modules and self.config.modules.cloud and len(self.config.modules.cloud):
                self.scanCloud()
            try:
                self.create_template()
            except:
                self.log(tag="ERROR", msg="Template Creation Failed")
        self.log(msg='', tag="Finished")

    # Excludes files in .git directories. Takes path of full path with filename
    def allowfile(self, path, allowDir=False):
        normpath = os.path.normpath(path)
        dirlist = normpath.split(os.sep)
        if (
            "node_modules" not in dirlist and
            ".git" not in dirlist and
            not os.path.islink(path) and
            (os.path.isfile(path) or allowDir)
        ):
            return True
        else:
            return False

    # Get recursive size of a directory
    def get_raw_size(self, start_path='.'):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(start_path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if (os.path.isfile(fp) and not os.path.islink(fp)):
                    total_size += os.path.getsize(fp)  # os.stat(fp).st_size
        return total_size

    # Simple algorithm for lines of code in a file
    def getloc(self, file):
        try:
            count = 0
            with open(file) as fp:
                for line in fp:
                    if line.strip():
                        count += 1
            return count
        except:
            return 0

    def checkDependency(self, command, name, kill=False) -> bool:
        which = shutil.which(command)
        if not which:
            self.log(msg=f"{name} is required but it's not installed.", tag=f"{name} status", display=False, timestamp=False)
            if kill:
                raise Exception(f"{name} is required but it's not installed.")
            return False
        else:
            self.log(msg=f"{name} is installed at {which}.", tag=f"{name} status", display=True, timestamp=False)
            return True

    def upload(self, file: str, route: str, source: str = '', isJSON=True):
        if not self.config.shouldUpload:
            self.log(
                msg='Skipping Uploads',
                tag=f"Skipping uploading {file} for {source} to {self.config.baseUrl}/{route}.",
                display=True
            )
            return True
        if not Path(file).exists():
            self.log(
                msg=f"File does not exist: {file}"
            )
            return False

        orig_route = route

        route = self.up(
            path_type=route,
            report=self.config.reportId,
            code=self.scanId,
            repo_name=source
        )

        with open(file, 'rb') as f:
            self.log(msg=f"{self.config.baseUrl}{route}", tag="Uploading to")

            if isJSON:
                headers = {
                    'Content-Type': 'application/json',
                    'accept': 'application/json'
                }
                response = requestx.post(self.config.baseUrl + route, data=f, headers=headers)
            else:
                files = {'logFile': f}
                response = requestx.post(self.config.baseUrl + route, files=files)

        if response.status_code == 200:
            self.log(
                msg='',
                tag=f"Successfully uploaded {file} for {source} to {self.config.baseUrl}{route}.",
                display=True
            )
            try:
                err_path_str = file[0:-5]+'.err'
                err_path = Path(err_path_str)

                if err_path.exists():
                    self.upload(
                        file=err_path_str,
                        route="err_"+orig_route,
                        source=source+" Error Logs",
                        isJSON=False
                    )
            except:
                pass
            return True
        else:
            self.log(
                msg=response.status_code,
                tag=f"Failed to upload {file} for {source} to {self.config.baseUrl}{route}",
                display=True
            )
            return False

    def formatGitHash(self, hash: str):
        hash = hash.replace("'", "").replace('"', "")
        message = std_exec(["git", "log", "-n1", "--pretty=format:%B", hash], self.log)
        author = std_exec(["git", "log", "-n1", "--pretty=format:%aN <%aE>", hash], self.log)
        commit = std_exec(["git", "log", "-n1", "--pretty=format:%H", hash], self.log)
        date = std_exec(["git", "log", "-n1", "--pretty=format:%aD", hash], self.log)
        signed = std_exec(["git", "show", "--format='%G?'", hash], self.log)
        if signed != 'N':
            signed = True
        else:
            signed = False
        merge = False
        merge1 = std_exec(["git", "show", hash], self.log)
        if merge1.startswith("Merge: "):
            merge = True
        returnVal = {
            "message": trimLineBreaks(message),
            "author": author,
            "commit": commit,
            "date": escapeChars(date),
            "signed": signed,
            "merge": merge
        }
        return returnVal

    def parseRepo(self, path: str, repo_name: str, branch: str = None):
        self.log(msg='parseRepo')

        if not self.config.dry:
            os.chdir(path)

        if branch is None:
            branch = "main"

        # Adding this for Windows support
        # Appears to fail with blank HEAD
        if self.config.runGit:
            std_exec(["git", "init"])

        if self.config.runGit and self.checkDependency("git", "Git"):
            try:
                if not self.config.dry:
                    subprocess.check_call(["git", f"--work-tree={path}", "checkout", branch])
            except subprocess.CalledProcessError:
                try:
                    if not self.config.dry:
                        subprocess.check_call(["git", "checkout", "master"])
                        branch = "master"
                except subprocess.CalledProcessError:
                    try:
                        cmd = "git for-each-ref --count=1 --sort=-committerdate refs/heads/ --format='%(refname:short)'"
                        # remove new lines and apostrophes from branch name.
                        branch = std_exec(cmd.split(" ")).replace("'", "").replace("\n", "")
                        subprocess.check_call(["git", "checkout", branch])
                    except subprocess.CalledProcessError:
                        if self.config.runGit:
                            raise Exception("Error checking out branch from git.")
                        else:
                            self.log("Error checking out branch from git.")

        # Git Stats
        git_output_file = os.path.join(self.config.output_dir, repo_name + ".git.log.json")
        if self.config.runGit:
            self.log(msg=repo_name, tag="Gathering source code statistics for", display=True)
            command = f'''git log \
                --since="{self.config.modules.code.git.start}" \
                --numstat \
                --format='%H' \
                {branch} --
            '''
            try:
                if not self.config.dry:
                    results = subprocess.run(command, shell=True, stdout=subprocess.PIPE)
                    log = results.stdout.decode()
            except subprocess.CalledProcessError:
                self.log("Error getting log from git")
                raise Exception("Error getting log from git.")

            if not self.config.dry:
                resultArr = log.split("\n")
                prevHash = ''
                filesArr = []
                finalArr = []

                for line in resultArr:
                    lineArr = line.split("\t")
                    if len(lineArr) > 1:
                        filesArr.append({
                            "insertions": lineArr[0],
                            "deletions": lineArr[1],
                            "path": lineArr[2]
                        })
                    else:
                        if len(lineArr) == 1 and lineArr[0] != '':
                            # Hit next file
                            if prevHash:
                                # Not first one
                                hashObj = self.formatGitHash(prevHash)
                                hashObj['paths'] = filesArr
                                finalArr.append(hashObj)
                                filesArr = []
                            prevHash = lineArr[0]

                self.log(msg=truncate(finalArr), tag=f"{repo_name} Git Stats")

                self.log(msg=git_output_file, display=True)

                with open(git_output_file, 'w') as f:
                    f.write(json.dumps(finalArr, indent=4))
                template_definition["gitlog"] = finalArr
                # End if not self.config.dry:

        if Path(git_output_file).exists():
            self.upload(
                file=git_output_file,
                route="git",
                source=repo_name
            )

        # Manual File Sizes and Info
        sizes_output_file = os.path.join(self.config.output_dir, repo_name + ".sizes.json")

        if not self.config.dry:
            self.log(msg=repo_name, tag="Gathering file sizes for", display=True)
            # Sizes for writing to output file
            # Intialize file list with "." as total size
            repo_size = self.get_raw_size(".")
            git_size = self.get_raw_size("./.git")

            # get sizes
            real_size = repo_size - git_size
            self.log(msg=repo_size, tag="Repo Size")
            self.log(msg=git_size, tag="Git Size")
            sizes = {
                "files": {
                    ".": {
                        "size": repo_size,
                        "loc": 0,
                        "ext": None,
                        "directory": True
                    }
                },
                "metadata": {
                    "env": machine,
                    "real_size": real_size,
                    "uname": system,
                    "branch": locals()['branch'] if "branch" in locals() else None
                }
            }

            # filelist for modernmetric
            filelist = []

            for filepath, subdirs, list in os.walk("."):
                for name in list:
                    fp = os.path.join(filepath, name)
                    extRe = re.search(r"^[^\.]*\.(.*)", name)
                    ext = extRe.group(1) if extRe else ''
                    if self.allowfile(path=fp):
                        if self.config.shouldManualFileScan:
                            file = {
                                "size": os.path.getsize(fp),
                                "loc": self.getloc(fp),
                                "ext": ext,  # os.path.splitext(name)[1],
                                "directory": False
                            }
                            sizes["files"][fp] = file
                        filelist.append({"name": name, "path": fp})

            with open(sizes_output_file, 'w') as f:
                f.write(json.dumps(sizes, indent=4))
            template_definition["current_dir_size"] = sizes["files"].pop(".")
            template_definition["sizes"] = sizes
            # End if not self.config.dry:

        self.upload(
            file=sizes_output_file,
            route="sizes",
            source=repo_name
        )

        # Run Modernmetric
        if self.config.runStats:
            stats_input_file = os.path.join(self.config.output_dir, repo_name + ".filelist.json")
            stats_output_file = os.path.join(self.config.output_dir, repo_name + ".stats.json")

            if not self.config.dry:
                self.log(msg=repo_name, tag="Analyzing repository with Modernmetric", display=True)
                with open(stats_input_file, 'w') as f:
                    f.write(json.dumps(filelist, indent=4))

                template_definition["filelist"] = filelist
                custom_args = [f"--file={stats_input_file}", f"--output={stats_output_file}"]
                modernmetric(custom_args)
                with open(stats_output_file, 'r') as f:
                    template_definition["stats"] = json.load(f)
            self.upload(
                file=stats_output_file,
                route="stats",
                source=repo_name
            )

        # Run SEMGrep
        if self.config.runScan:
            if system.lower() == 'windows':
                self.log("""
                Windows does not support Semgrep.
                Please see the open issues here:
                https://github.com/returntocorp/semgrep/issues/1330
                         """)
            if self.checkDependency('semgrep', "Semgrep"):
                findings_output_file = os.path.join(self.config.output_dir, repo_name + ".findings.json")
                findings_error_file = os.path.join(self.config.output_dir, repo_name + ".findings.err")
                if not self.config.dry:
                    self.log(msg=repo_name, tag="Scanning repository", display=True)
                    try:
                        with open(findings_error_file, 'a') as e:
                            subprocess.check_call([
                                "semgrep",
                                "scan",
                                "--config",
                                "auto",
                                "--json",
                                "-o",
                                findings_output_file,
                            ], stderr=e,)
                    except subprocess.CalledProcessError as e:
                        output = e.output
                        self.log(msg=output, tag="Scanning repository return", display=True)
                try:
                    with open(findings_output_file) as f:
                        findings = json.load(f)

                    # This is on purpose. If you try to read same pointer
                    # twice, it dies.
                    with open(findings_output_file) as f:
                        original_findings = json.load(f)

                    if self.config.truncate_findings:
                        # Exclusions are set to exclude fields that are not code
                        truncation_exclusion = [
                            "cwe",
                            "owasp",
                            "path",
                            "check_id",
                            "license",
                            "fingerprint",
                            "message",
                            "references",
                            "url",
                            "source",
                            "severity"
                        ]
                        self.log(
                            tag="TRUNCATING",
                            msg=f"excluding: {truncation_exclusion}"
                        )
                        try:
                            findings = truncate_children(
                                findings,
                                self.log,
                                excludes=truncation_exclusion,
                                max_length=self.config.truncate_findings_length
                            )
                        except Exception as e:
                            self.log(tag="ERROR", msg="Error in Truncation")
                            self.log(e)
                            self.log(
                                json.dumps(
                                    original_findings,
                                    indent=4,
                                    sort_keys=True
                                )
                            )
                    with open(findings_output_file, "w") as f2:
                        f2.write(json.dumps(
                            findings, indent=4, sort_keys=True
                        ))
                    template_definition["gitfindings"] = findings
                except Exception as e:
                    if not self.config.dry:
                        raise e
                    else:
                        self.log(
                            msg=f'''
                                Attempted to format/truncate non existent file
                                {findings_output_file}
                            '''
                        )
                self.upload(
                    file=findings_output_file,
                    route="findings",
                    source=repo_name
                )

        # ##### Scan Dependencies ######
        if self.config.runDependencies:
            dependencies_output_file = os.path.join(self.config.output_dir, repo_name + ".dependencies.json")
            self.log(msg=repo_name, tag="Scanning dependencies", display=True)
            if not self.config.dry:
                dependencies_output_file = dependency_walk(output_file=dependencies_output_file, logger=self.log)
                with open(dependencies_output_file, "r") as f:
                    template_definition["dependencies"] = json.load(f)
            self.log(msg=dependencies_output_file, tag="Dependency File", display=False)
            self.upload(
                file=dependencies_output_file,
                route="dependencies",
                source=repo_name
            )

    def preflight(self):
        # If the 'dry' configuration is set, skip the preflight checks
        if self.config.dry:
            return

        # Loop over all remote repositories from config file
        print("\n\n\nChecking your system's compatibility with the scan configuration:\n")
        if 'repos' in self.config.config:
            repos = self.config.config["repos"]
            if repos:
                for repo_url in [r for r in repos if len(r) > 0]:       # ignore blank lines from server
                    match = re.search(r"([^/]*\.git.*)", repo_url)
                    if match:
                        repo_name = match.group(1)
                    else:
                        repo_name = repo_url.rsplit('/', 1)[-1]
                    if "@" in repo_name and re.search(r"^.*@.*\..*:", repo_url):
                        repo_url = "@".join(repo_url.split("@")[0:2])
                    elif "@" in repo_name:
                        repo_url = repo_url.split("@")[0]
                    try:
                        subprocess.check_output(["git", "ls-remote", repo_url])
                        self.log(tag="Repository access confirmed", msg=repo_url, display=True, timestamp=False)
                    except subprocess.CalledProcessError:
                        self.log(msg=repo_url, tag="Unable to access", display=True, timestamp=False)
                        self.log(msg=repo_url, tag="Repository will not be scanned", display=True, timestamp=False)

        cloud_config = self.config.modules.cloud
        if cloud_config is not None:
            for provider in cloud_config:
                try:
                    if provider.provider == "aws" and self.checkDependency("aws", "AWS Command-line tool"):
                        account_id = str(provider.account).replace('-', '')
                        if find_profile(account_id, self.log) is None:
                            self.log(tag=f"No matching AWS CLI profiles found for {provider.account}", msg="Account can't be scanned.", display=True, timestamp=False)
                        else:
                            self.log(tag="AWS account access confirmed", msg=account_id, display=True, timestamp=False)
                    if provider.provider == "azure" and self.checkDependency("az", "Azure Command-line tool"):
                        pass
                    if provider.provider == "gcp" and self.checkDependency("gcloud", "Google Command-line tool"):
                        pass
                except:
                    self.log(msg=f"Unable to access {provider.provider} {provider.account}", tag="Unable to access", display=True, timestamp=False)

        resp = repeat_boolean_prompt(
            "\nWould you like to proceed with the scan?",
            logger=print,
            default_val=True
        )

        if resp:
            self.log(msg="Proceeding")
        else:
            self.log(tag="Exiting now", msg="", display=True)
            exit(0)

    # ##### Scan Repos ######
    def scanRepos(self):
        # Loop over all remote repositories from config file
        if 'repos' in self.config.config:
            repos = self.config.config["repos"]
            if repos:
                for repo_url in [r for r in repos if len(r) > 0]:       # ignore blank lines from server
                    repo_info = get_repo_name_url_and_branch(repo_url)
                    repo_name = repo_info["repo_name"]
                    repo_url = repo_info["repo_url"]
                    branch = repo_info["branch"]
                    self.log(msg=repo_name, tag="Processing", display=True)
                    self.log(msg=repo_url, tag="URL", display=True)
                    self.log(msg=branch, tag="Branch Specified", display=True)
                    try:
                        if not self.config.dry:
                            os.makedirs(temp_dir)
                    except:
                        self.log(tag="Directory exists:", msg=temp_dir, display=True)
                        try:
                            shutil.rmtree(temp_dir)
                            os.makedirs(temp_dir)
                        except Exception as e:
                            self.log(tag=f"Failed to delete {temp_dir}", msg=e, display=True)
                            continue

                    self.log(msg=repo_url, tag="Repo URL")
                    self.log(msg=temp_dir, tag="Temp Directory")
                    if not self.config.dry and self.config.runGit:
                        try:
                            subprocess.check_output(["git", "clone", repo_url, temp_dir])
                        except subprocess.CalledProcessError:
                            self.log(msg=repo_url, tag="Failed to clone", display=True)
                            exit(1)
                            continue

                        self.log(msg=repo_url, tag="Successfully cloned", display=True)
                    try:
                        self.parseRepo(temp_dir, repo_name, branch)
                    except Exception as e:
                        self.log(msg=str(e), tag="parseRepo Error Caught")
                        self.log(tag="", msg=traceback.format_exc())

                    os.chdir(curr_dir)
                    if not self.config.dry and self.config.delete_temp:
                        shutil.rmtree(temp_dir)
            else:
                self.log(msg='', tag="No remote repos", display=True)

        else:
            self.log(msg='', tag="No remote repos", display=True)

        # Loop over all local repositories from config file
        if 'local_repos' in self.config.config:
            localrepos = self.config.config['local_repos']
            if localrepos:
                for repo_path in localrepos:
                    a = Path(repo_path).absolute()
                    match = re.search(r"([^/]*\.git.*)", str(a))
                    if match:
                        repo_name = match.group(1)
                    else:
                        repo_name = os.path.basename(os.path.normpath(repo_path))
                    self.parseRepo(repo_path, repo_name)
            else:
                self.log(msg='', tag="No local repos", display=True)
        else:
            self.log(msg='', tag="No local repos", display=True)

        self.log(msg='', tag="Finished repo scans")

    # ##### Scan Cloud ######
    def scanCloud(self):
        self.log(msg='', tag="Doing cloud scan", display=True)
        cloud_config = self.config.modules.cloud
        self.log(msg=cloud_config, tag='Cloud Config')

        if cloud_config is None:
            return
        for provider in cloud_config:
            try:
                # Check if AWS-CLI is installed
                if provider.provider == "aws" and self.checkDependency("aws", "AWS Command-line tool"):
                    account_id = str(provider.account).replace('-', '')
                    aws_cost_file = runAws(
                        targeted_account=account_id,
                        start=provider.start,
                        end=provider.end,
                        profile=provider.profile,
                        path_to_output=self.config.output_dir,
                        log=self.log,
                        dry=self.config.dry
                    )
                    if aws_cost_file is None:
                        self.log(msg="Error processing AWS costs", tag=account_id)
                    else:
                        self.log(msg=aws_cost_file, tag="AWS Costs")
                        self.upload(
                            file=aws_cost_file,
                            route="costs",
                            source="AWS"
                        )
                    aws_instance_file = get_aws_instances(
                        sub_id=account_id,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry
                    )
                    if aws_instance_file is None:
                        self.log(msg="Error processing AWS instances", tag=account_id)
                    else:
                        self.log(msg=aws_instance_file, tag="AWS Instances")
                        self.upload(
                            file=aws_instance_file,
                            route="instances",
                            source="AWS"
                        )
                    aws_utilization_file = os.path.join(
                        self.config.output_dir,
                        f'aws-instances-{account_id}-utilization.json'
                    )
                    if Path(aws_utilization_file).is_file():
                        self.upload(
                            file=aws_utilization_file,
                            route="utilization",
                            source="AWS"
                        )
                    aws_block_file = get_aws_blocks(
                        sub_id=account_id,
                        path_to_output=self.config.output_dir,
                        log=self.log,
                        dry=self.config.dry
                    )
                    if aws_block_file is None:
                        self.log(msg="Error processing AWS blocks", tag=account_id)
                    else:
                        self.log(msg=aws_block_file, tag="AWS Storage")
                        self.upload(
                            file=aws_block_file,
                            route="storage",
                            source="AWS"
                        )

                # Check if Azure CLI is installed
                if provider.provider == "azure" and self.checkDependency("az", "Azure Command-line tool"):
                    azure_cost_file = runAzure(
                        subscription_id=provider.account,
                        start=provider.start,
                        end=provider.end,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry
                    )
                    if azure_cost_file is None:
                        self.log(msg="Error processing Azure costs", tag=provider.account)
                    else:
                        self.log(msg=azure_cost_file, tag="Azure Costs")
                        self.upload(
                            file=azure_cost_file,
                            route="costs",
                            source="Azure"
                        )
                    azure_instance_file = get_az_instances(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry,
                        log=self.log
                    )
                    if azure_instance_file is None:
                        self.log(msg="Error processing Azure instances", tag=provider.account)
                    else:
                        self.log(msg=azure_instance_file, tag="Azure instances")
                        self.upload(
                            file=azure_instance_file,
                            route="instances",
                            source="Azure"
                        )
                    azure_utilization_file = os.path.join(
                        self.config.output_dir,
                        f'azure-instances-{provider.account}-utilization.json'
                    )
                    if Path(azure_utilization_file).is_file():
                        self.upload(
                            file=azure_utilization_file,
                            route="utilization",
                            source="AWS"
                        )
                    azure_block_file = get_az_blocks(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry
                    )
                    if azure_block_file is None:
                        self.log(msg="Error processing Azure blocks", tag=provider.account)
                    else:
                        self.log(msg=azure_block_file, tag="Azure Storage")
                        self.upload(
                            file=azure_block_file,
                            route="storage",
                            source="Azure"
                        )

                if provider.provider == "gcp" and self.checkDependency("gcloud", "Google Command-line tool"):
                    gcp_instance_file = get_gcp_instances(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry
                    )
                    if gcp_instance_file is None:
                        self.log(msg="Error processing GCP instances", tag=provider.account)
                    else:
                        self.log(msg=gcp_instance_file, tag="GCP instances")
                        self.upload(
                            file=gcp_instance_file,
                            route="instances",
                            source="GCP"
                        )
                    gcp_utilization_file = os.path.join(
                        self.config.output_dir,
                        f'gcp-instances-{provider.account}-utilization.json'
                    )
                    if Path(gcp_utilization_file).is_file():
                        self.upload(
                            file=gcp_utilization_file,
                            route="utilization",
                            source="AWS"
                        )
                    gcp_block_file = get_gcp_blocks(
                        sub_id=provider.account,
                        path_to_output=self.config.output_dir,
                        dry=self.config.dry
                    )
                    if gcp_block_file is None:
                        self.log(msg="Error processing GCP blocks", tag=provider.account)
                    else:
                        self.log(msg=gcp_block_file, tag="GCP Storage")
                        self.upload(
                            file=gcp_block_file,
                            route="storage",
                            source="GCP"
                        )
            except Exception as e:
                self.log(tag="ERROR", msg="Error processing provider", display=True)
                self.log(
                    tag="ERROR PROVIDER",
                    msg=str(provider),
                    display=True
                )
                self.log(tag="ERROR", msg=e, display=True)
                self.log(
                    tag="ERROR STACK",
                    msg=traceback.format_exc()
                )


def main():
    agent = Agent()
    try:
        agent.preflight()
        agent.scan()
    except Exception as e:
        agent.log(msg=str(e), tag="Main Scan Error Caught")
        if agent.config.upload_logs:
            agent.upload(route="logs", file=agent.config.output_dir+"/log.txt")
        raise e
    if agent.config.shouldUpload or agent.config.upload_logs:
        agent.upload(route="logs", file=agent.config.output_dir+"/log.txt", source='logs', isJSON=False)
    if agent.config.upload_logs:
        new_folder_name = (
            str(today.year) + str(today.month) + str(today.day)
        )
        d = agent.directory
        os.makedirs(f'{d}/{new_folder_name}/', exist_ok=True)
        new_file_name = str(uuid4())+".txt"
        fp = f'{d}/{new_folder_name}/{new_file_name}'
        shutil.copy2(agent.config.output_dir+"/log.txt", fp)
        os.unlink(agent.config.output_dir+"/log.txt")
        print(f"""The log for this run has moved to:
              {d}/{new_folder_name}/{new_file_name}""")

    # We only do this if you have a remote config but didn't upload
    if not agent.config.shouldUpload and agent.config.is_original_path_remote():
        print("\n\n\nIMPORTANT: To upload results from this location please run")
        print(f"verinfast -c {agent.config.cfg_path} -o {agent.config.output_dir} --should_upload --dry")


if __name__ == "__main__":
    main()
# For test runs from commandline. Comment out before packaging. # main()



from verinfast.config import UploadConfig


class Uploader:
    def __init__(self, config: UploadConfig):
        self.config = config

    def make_upload_path(
            self,
            path_type: str,
            report: int | str,
            code: int | str = None,
            repo_name: str = None,
    ) -> str:
        """make_upload_path
        make_upload_path is a convenience function for abstracting
        the url paths from the scanning logic.

        Args:
            path_type (str) : specifies the scan action triggering the upload
            report (str | int): specifies the identifier the server uses for this scan  # noqa: E501
            code (int | str) :
        """

        code_sep = self.config.code_separator if self.config.code_separator else ''  # noqa: E501
        cost_sep = self.config.cost_separator if self.config.cost_separator else ''  # noqa: E501

        paths = {
            "git": f"{report}{code_sep}/{code}/{repo_name}/git",
            "sizes": f"{report}{code_sep}/{code}/{repo_name}/sizes",
            "pygount": f"{report}{code_sep}/{code}/{repo_name}/pygount",
            "stats": f"{report}{code_sep}/{code}/{repo_name}/stats",
            "findings": f"{report}{code_sep}/{code}/{repo_name}/findings",
            "dependencies": f"{report}{code_sep}/{code}/{repo_name}/dependencies",  # noqa: E501
            "costs": f"{report}{cost_sep}/costs",
            "instances": f"{report}{cost_sep}/instances",
            "utilization": f"{report}{cost_sep}/instance_utilization",
            "storage": f"{report}{cost_sep}/storage",
            "scan_id": f"{report}{code_sep}",
            "logs": f"{report}/agent_logs",
            "err_stats": f"{report}/agent_err/stats_err",
            "err_findings": f"{report}/agent_err/findings_err"
        }

        if report is None:
            raise Exception("Invocation Error: must supply either a report ID or UUID")  # noqa: E501

        requires_code = [
            "git",
            "sizes",
            "pygount",
            "stats",
            "findings",
            "dependencies"
        ]

        if path_type in requires_code and code is None:
            raise Exception(f"Invocation Error: type {path_type} requires code")  # noqa: E501

        if path_type in requires_code and repo_name is None:
            raise Exception(f"Invocation Error: type {path_type} requires repo_name")  # noqa: E501

        return_path = paths[path_type]

        # logs never gets uuid prefix
        if self.config.uuid and path_type != "logs":
            return_path = "uuid/" + return_path

        if self.config.prefix is not None:
            return_path = self.config.prefix + return_path

        return return_path



import datetime
import json
import os
from typing import List

from azure.identity import DefaultAzureCredential
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.monitor.query import MetricsQueryClient, MetricAggregationType


from verinfast.cloud.cloud_dataclass import \
    Utilization_Datapoint as Datapoint,  \
    Utilization_Datum as Datum


metric_name = "Percentage CPU"
d = datetime.timedelta(
                days=30,
                seconds=0,
                microseconds=0,
                milliseconds=0,
                minutes=0,
                hours=0,
                weeks=0
            )

metric_identifiers = {"cpu": metric_name}
aggregations = [
    MetricAggregationType.MINIMUM,
    MetricAggregationType.AVERAGE,
    MetricAggregationType.MAXIMUM
]


def get_metrics_for_instance(
        metrics_client: MetricsQueryClient,
        instance_id: str,
        instance_name: str
        ) -> List[Datum]:
    print("Get Metrics for Instance")
    data = []
    try:
        o = metrics_client.query_resource(
            resource_uri=instance_id,
            interval=datetime.timedelta(hours=1),
            aggregations=aggregations,
            metric_names=[metric_name],
            timespan=d,

        )
        for dp in o.metrics[0].timeseries[0].data:
            n = Datapoint(
                Minimum=dp.minimum,
                Maximum=dp.maximum,
                Average=dp.average
            )
            datum = Datum(
                Timestamp=dp.timestamp.timestamp(),
                cpu=n
            )
            data.append(datum)

    except Exception as e:
        print(e)
        pass

    return data


def get_instances(
        sub_id: str,
        path_to_output: str = "./",
        dry=False,
        log=None
        ):

    if not dry:
        my_instances = []
        metrics = []
        client = ComputeManagementClient(
                credential=DefaultAzureCredential(),
                subscription_id=sub_id
            )
        metrics_client = MetricsQueryClient(
                credential=DefaultAzureCredential(),
                subscription_id=sub_id
            )
        networkClient = NetworkManagementClient(
                credential=DefaultAzureCredential(),
                subscription_id=sub_id
            )

        res = client.virtual_machines.list_all()

        for vm in res:
            try:
                s = vm.storage_profile.os_disk.managed_disk.id
                tgt = s.split('resourceGroups/')[1]
                tgt = tgt.split('/')[0]
                name = vm.name
                location = vm.location
                hw = vm.hardware_profile.vm_size
                iv = client.virtual_machines.instance_view(
                        vm_name=name,
                        resource_group_name=tgt
                    )
                status = iv.statuses[-1]
                network_interface = vm.network_profile.network_interfaces[0]
                nic_name = network_interface.id.split("networkInterfaces/")[1]
                nics = networkClient.network_interface_ip_configurations.list(
                        resource_group_name=tgt,
                        network_interface_name=nic_name
                    )
                subnet = ""
                public_ip = "n/a"
                for nic in nics:
                    subnet = nic.subnet.name
                    vnet_name = nic.subnet.id.split('/')[-3]
                    if (
                        hasattr(nic, "public_ip_address") and
                        hasattr(nic.public_ip_address, "ip_address")
                    ):
                        public_ip_name = nic.public_ip_address.id.split('/')[-1]  # noqa: E501
                        pi = networkClient.public_ip_addresses.get(tgt, public_ip_name)  # noqa: E501
                        public_ip = pi.ip_address
                architecture = "x86_64"

                if "D2p" in hw or "E2p" in hw:
                    architecture = "Arm"

                my_instance = {
                    "id": vm.id,
                    "name": name,
                    "type": hw,
                    "state": status.display_status,
                    "zone": vm.zones[0] if vm.zones else None,
                    "region": location,
                    "subnet": subnet,
                    "architecture": architecture,
                    "publicIp": public_ip,
                    "vpc": vnet_name
                }
            except Exception as e:
                if log:
                    log(
                        tag="Azure Get Instance Error",
                        msg=str(e)
                    )
                continue
            my_instances.append(my_instance)
            m = get_metrics_for_instance(
                metrics_client=metrics_client,
                instance_id=vm.id,
                instance_name=name
            )
            d = {
                "id": vm.id,
                "metrics": [metric.dict for metric in m]
            }
            metrics.append(d)
        upload = {
                    "metadata": {
                        "provider": "azure",
                        "account": str(sub_id)
                    },
                    "data": my_instances
                }
    # End dry block

    azure_output_file = os.path.join(
        path_to_output,
        f'az-instances-{sub_id}.json'
    )
    if not dry:
        with open(azure_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))

        upload['data'] = metrics
        with open(azure_output_file[:-5]+"-utilization.json", "w") as outfile2:
            outfile2.write(json.dumps(upload, indent=4))
    return azure_output_file

# Test code
# print(get_instances(sub_id="80dc7a6b-df94-44be-a235-7e7ade335a3c"))



import csv
import io
import json
import subprocess

# from pathlib import Path
# from typing import List

# Important notes: Ruby will install at most one version of a gem, and will fail if versions conflict.


from gemfileparser import GemfileParser, Dependency

from verinfast.dependencies.walkers.classes import Walker, Entry


class GemWalker(Walker):
    def parse(self, file: str, expand=False):
        command = f"gem install -r --explain --no-user-install --install-dir gems -g {file}"
        try:
            results = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.loggerFunc(msg=f'gem install results: {results.stderr.decode()}')

            log = results.stdout.decode()
            self.real_dependencies = {}
            for line in log.splitlines():
                line = line.strip()
                if line == "Gems to install:":
                    pass
                else:
                    split_position = line.rindex('-')
                    gem_name = line[0:split_position]
                    gem_version = line[split_position+1:]
                    self.real_dependencies[gem_name] = gem_version
        except:  # noqa:E722
            self.loggerFunc(msg=f"Failed to parse Gemfile {file}", display=True)
            with open(file, "r") as manifest:
                self.loggerFunc(msg=file)
                contents = manifest.read()
                self.loggerFunc(msg=contents+"\n\n\n")
            return

        # Yes this is necessary
        with open(file, "r") as manifest:
            self.lines = manifest.readlines()
            do_nest = 0
            first_source = False
            sources = []
            source = None
            deps = []
            for line in self.lines:
                line = GemfileParser.preprocess(line)
                if len(line) > 0:
                    if line.endswith(" do"):
                        do_nest += 1
                    if line.startswith("source"):
                        source = line.split()[1].replace('"', "").replace("'", "")
                        sources.append({
                            "source": source,
                            "layer": do_nest
                        })
                        first_source = True
                    if not first_source and not line.startswith("ruby"):
                        self.loggerFunc(msg=manifest.read() + '\n\n\n')
                        self.loggerFunc(msg=f"Invalid gemfile {file}")
                        return
                    if line == 'end':
                        if len(sources) == 0:
                            self.loggerFunc(msg=manifest.read() + '\n\n\n')
                            self.loggerFunc(msg=f"Invalid gemfile {file}")
                            return
                        last_source = sources[-1]
                        if do_nest == last_source["layer"]:
                            sources.pop()
                            if len(sources) > 0:
                                source = sources[-1]["source"]
                            else:
                                source = None
                        do_nest -= 1

                    if line.startswith("gem"):
                        linefile = io.StringIO(line)
                        for line2 in csv.reader(linefile, delimiter=","):
                            column_list = []
                            for column in line2:
                                stripped_column = (
                                    column.replace("'", "")
                                    .replace('"', "")
                                    .replace("%q<", "")
                                    .replace("(", "")
                                    .replace(")", "")
                                    .replace("[", "")
                                    .replace("]", "")
                                    .replace(".freeze", "")
                                    .strip()
                                )
                                column_list.append(stripped_column)

                                dep = Dependency()
                                # dep.group = self.current_group
                                # dep.parent.append(self.appname)
                                for column in column_list:
                                    # Check for a match in each regex and assign to
                                    # corresponding variables
                                    for criteria, criteria_regex in GemfileParser.gemfile_regexes.items():
                                        match = criteria_regex.match(column)
                                        if match:
                                            if criteria == "requirement":
                                                dep.requirement.append(match.group(criteria))
                                            else:
                                                setattr(dep, criteria, match.group(criteria))
                                            break
                                if dep.name.startswith("gem "):
                                    dep.name = dep.name[4:].strip()
                                if not dep.requirement:
                                    dep.requirement.append("*")
                                if dep.source == '' or dep.source is None:
                                    dep.source = source
                                if dep.source == '' or dep.source is None:
                                    dep.source = "https://rubygems.org"
                                if dep.source == "https://rubygems.org":
                                    self.get_license(dep.name)
                                deps.append(dep)
                    else:
                        pass

        for dep in deps:
            license = self.get_license(getattr(dep, "name"))
            e = Entry(
                        name=dep.name,
                        specifier=" ".join(dep.requirement),
                        source=dep.source,
                        license=license,
                        required_by=file
                    )
            self.entries.append(e)

        # Second order dependencies
        for dep in self.real_dependencies:
            if dep not in [n.name for n in deps]:
                e = Entry(
                    name=dep,
                    specifier="=="+self.real_dependencies[dep],
                    source="gemfile child",
                    license=self.get_license(dep)
                )
                self.entries.append(e)

    def get_license(self, name: str) -> str:
        if name not in self.real_dependencies:
            return ""
        version = self.real_dependencies[name]
        license = ""
        try:
            resp = self.getUrl(f'https://rubygems.org/api/v2/rubygems/{name}/versions/{version}.json')
            resp = json.loads(resp)
            license = " ".join(resp["licenses"]) if resp["licenses"] is not None else ""
        except Exception as e2:
            self.log(
                f"License not found for: {name}@{version}",
                display=False
            )
            self.log(e2, display=False)
        return license

    def expand(self, file):
        raise Exception("No expansion for this Walker")



import json
from pathlib import Path
from verinfast.utils.utils import get_repo_name_url_and_branch

file_path = Path(__file__)
test_folder = file_path.parent.absolute()
fixtures_dir = test_folder.joinpath("fixtures").absolute()
url_list_path = str(
    fixtures_dir.joinpath('url_support/test_repo_urls.json').absolute()
)
with open(url_list_path, "r") as f:
    url_list = json.load(f)


def test_repo_urls():
    for url in url_list:
        repo_info = get_repo_name_url_and_branch(url["url"])
        repo_name = repo_info["repo_name"]
        repo_url = repo_info["repo_url"]
        branch = repo_info["branch"]
        assert repo_name == url["expected_name"], (
            f"Expected name: {url['expected_name']}, "
            f"Got: {repo_name} for {url['test']}"
        )
        assert repo_url == url["expected_url"], (
            f"Expected repo_url: {url['expected_url']}, "
            f"Got: {repo_url} for {url['test']}"
        )
        assert branch == url["expected_branch"], (
            f"Expected branch: {url['expected_branch']}, "
            f"Got: {branch} for {url['test']}"
        )
        print(f"Test passed for {url['test']}")
        print(f"Name: {repo_name}, URL: {repo_url}, Branch: {branch}")



regions = [
    "us-east-2",
    "us-east-1",
    "us-west-1",
    "us-west-2",
    "af-south-1",
    "ap-east-1",
    "ap-south-2",
    "ap-southeast-3",
    "ap-southeast-4",
    "ap-south-1",
    "ap-northeast-3",
    "ap-northeast-2",
    "ap-southeast-1",
    "ap-southeast-2",
    "ap-northeast-1",
    "ca-central-1",
    "eu-central-1",
    "eu-west-1",
    "eu-west-2",
    "eu-south-1",
    "eu-west-3",
    "eu-south-2",
    "eu-north-1",
    "eu-central-2",
    "me-south-1",
    "me-central-1",
    "sa-east-1"
]



import json
import os
import subprocess

from verinfast.utils.utils import DebugLog
from verinfast.cloud.aws.get_profile import find_profile
debugLog = DebugLog(os.getcwd())


def runAws(targeted_account, start, end, path_to_output,
           profile=None, log=debugLog.log, dry=False):

    def _get_costs_and_usage(profile: str, aws_output_file: str):
        cmd = f'''
            aws ce get-cost-and-usage \
            --time-period Start={start},End={end} \
            --granularity=DAILY \
            --metrics "BlendedCost" \
            --group-by Type=DIMENSION,Key=SERVICE \
            --profile="{profile}" \
            --output=json | cat
        '''

        try:
            results = subprocess.run(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                check=True
            )

        except subprocess.CalledProcessError:
            log(msg="Error getting data from AWS CLI get-cost-and-usage",
                tag="AWS CLI")
            raise Exception("Error getting aws cli data.")

        text = results.stdout.decode()
        if text is None or isinstance(text, str) is False:
            log(msg="No data returned from AWS CLI get-cost-and-usage",
                tag="AWS CLI")
            return None

        obj = json.loads(text)
        results_by_time = obj["ResultsByTime"]
        charges = []
        for charge in results_by_time:
            if charge["Groups"]:
                for group in charge["Groups"]:
                    newCharge = {
                        "Date": charge["TimePeriod"]["Start"],
                        "Group": group["Keys"][0],
                        "Cost": group["Metrics"]["BlendedCost"]["Amount"],
                        "Currency": group["Metrics"]["BlendedCost"]["Unit"]
                    }
                    charges.append(newCharge)
        upload = {
            "metadata": {
                "provider": "aws",
                "account": str(targeted_account)
            },
            "data": charges
        }

        with open(aws_output_file, 'w') as outfile:
            outfile.write(json.dumps(upload, indent=4))
        return aws_output_file

    if profile is None:
        profile = find_profile(targeted_account=targeted_account, log=log)

    if profile is None:
        log(msg="No matching profiles found",
            tag=targeted_account)
        return None

    output_file = os.path.join(
        path_to_output,
        f'aws-cost-{targeted_account}.json'
    )
    if not dry:
        _get_costs_and_usage(profile, output_file)

    return output_file



zones = [
    "asia-east1-a",
    "asia-east1-b",
    "asia-east1-c",
    "asia-east2-a",
    "asia-east2-b",
    "asia-east2-c",
    "asia-northeast1-a",
    "asia-northeast1-b",
    "asia-northeast1-c",
    "asia-northeast2-a",
    "asia-northeast2-b",
    "asia-northeast2-c",
    "asia-northeast3-a",
    "asia-northeast3-b",
    "asia-northeast3-c",
    "asia-south1-a",
    "asia-south1-b",
    "asia-south1-c",
    "asia-south2-a",
    "asia-south2-b",
    "asia-south2-c",
    "asia-southeast1-a",
    "asia-southeast1-b",
    "asia-southeast1-c",
    "asia-southeast2-a",
    "asia-southeast2-b",
    "asia-southeast2-c",
    "australia-southeast1-a",
    "australia-southeast1-b",
    "australia-southeast1-c",
    "australia-southeast2-a",
    "australia-southeast2-b",
    "australia-southeast2-c",
    "europe-central2-a",
    "europe-central2-b",
    "europe-central2-c",
    "europe-north1-a",
    "europe-north1-b",
    "europe-north1-c",
    "europe-southwest1-a",
    "europe-southwest1-b",
    "europe-southwest1-c",
    "europe-west1-b",
    "europe-west1-c",
    "europe-west1-d",
    "europe-west12-a",
    "europe-west12-b",
    "europe-west12-c",
    "europe-west2-a",
    "europe-west2-b",
    "europe-west2-c",
    "europe-west3-a",
    "europe-west3-b",
    "europe-west3-c",
    "europe-west4-a",
    "europe-west4-b",
    "europe-west4-c",
    "europe-west6-a",
    "europe-west6-b",
    "europe-west6-c",
    "europe-west8-a",
    "europe-west8-b",
    "europe-west8-c",
    "europe-west9-a",
    "europe-west9-b",
    "europe-west9-c",
    "me-central1-a",
    "me-central1-b",
    "me-central1-c",
    "me-west1-a",
    "me-west1-b",
    "me-west1-c",
    "northamerica-northeast1-a",
    "northamerica-northeast1-b",
    "northamerica-northeast1-c",
    "northamerica-northeast2-a",
    "northamerica-northeast2-b",
    "northamerica-northeast2-c",
    "southamerica-east1-a",
    "southamerica-east1-b",
    "southamerica-east1-c",
    "southamerica-west1-a",
    "southamerica-west1-b",
    "southamerica-west1-c",
    "us-central1-a",
    "us-central1-b",
    "us-central1-c",
    "us-central1-f",
    "us-east1-b",
    "us-east1-c",
    "us-east1-d",
    "us-east4-a",
    "us-east4-b",
    "us-east4-c",
    "us-east5-a",
    "us-east5-b",
    "us-east5-c",
    "us-south1-a",
    "us-south1-b",
    "us-south1-c",
    "us-west1-a",
    "us-west1-b",
    "us-west1-c",
    "us-west2-a",
    "us-west2-b",
    "us-west2-c",
    "us-west3-a",
    "us-west3-b",
    "us-west3-c",
    "us-west4-a",
    "us-west4-b",
    "us-west4-c"
]



import json
from unittest.mock import patch
import os
from pathlib import Path
from verinfast.agent import Agent


@patch('verinfast.user.__get_input__', return_value='y')
def test_local_scan(self):
    file_path = Path(__file__)
    test_folder = file_path.parent
    output_path = test_folder.joinpath("results").joinpath("tsx_test.stats.json")  # noqa: E501
    tsx_path = test_folder.joinpath("tsx_test")

    agent = Agent()
    agent.config.runGit = False
    agent.config.dry = False
    agent.config.runPygount = False
    agent.config.runScan = False
    agent.config.runSizes = False
    agent.config.runStats = True
    agent.config.output_dir = test_folder.joinpath("results")
    os.makedirs(agent.config.output_dir, exist_ok=True)
    agent.config.config["local_repos"] = [tsx_path]

    agent.scanRepos()

    with open(output_path) as output_file:
        output = json.load(output_file)
        assert output is not None

    return None



import os
from unittest.mock import patch
from pathlib import Path
import shutil

from verinfast.agent import Agent
from verinfast.config import Config
from verinfast.utils.utils import DebugLog

file_path = Path(__file__)
test_folder = file_path.parent.absolute()
results_dir = test_folder.joinpath("results").absolute()
str_path = str(test_folder.joinpath('str_conf.yaml').absolute())


@patch('verinfast.user.__get_input__', return_value='y')
def test_no_config(self):
    try:
        shutil.rmtree(results_dir)
    except Exception as e:
        print(e)
        pass
    os.makedirs(results_dir, exist_ok=True)
    agent = Agent()
    config = Config(str_path)
    config.output_dir = results_dir
    print(agent.config.output_dir)
    agent.config = config
    assert agent.config.use_uuid is True
    agent.config.shouldUpload = True
    agent.debug = DebugLog(path=agent.config.output_dir, debug=False)
    agent.log = agent.debug.log
    print(agent.debug.logFile)
    agent.scan()
    assert Path(results_dir).exists()
    assert results_dir.joinpath("results.html").exists()



import os
import datetime


# Get the version from the git tag, and write to VERSION.
ref = None
if "GITHUB_REF" in os.environ:
    ref = os.environ["GITHUB_REF"]

if ref and ref is not None and ref.startswith("refs/tags/"):
    version = ref.replace("refs/tags/", "")
else:
    version = datetime.datetime.now().strftime("%Y.%m.%d%H%M%S")

print(version)

with open("VERSION", "w") as f:
    f.write(version)

with open("VERSION.py", "w") as f:
    f.write("__version__ = '" + str(version) + "'")



from unittest.mock import patch
from pathlib import Path

from verinfast.agent import Agent
from verinfast.config import Config


def print_cloud(config_file: str):
    file_path = Path(__file__)
    test_folder = file_path.parent.absolute()
    agent = Agent()
    config = Config()
    config.cfg_path = str(test_folder.joinpath(config_file).absolute())
    config.__init__()
    # Test not overwritten
    assert config.cfg_path == str(test_folder.joinpath(config_file).absolute())  # noqa: E501
    assert config.modules.cloud
    agent.config = config
    for provider in agent.config.modules.cloud:
        print(provider)


@patch('verinfast.user.__get_input__', return_value='y')
def test_gcp(self):
    config_file = "gcp_conf.yaml"
    print_cloud(config_file=config_file)
    assert True


@patch('verinfast.user.__get_input__', return_value='y')
def test_aws(self):
    config_file = "aws_conf.yaml"
    print_cloud(config_file=config_file)
    assert True


@patch('verinfast.user.__get_input__', return_value='y')
def test_aws_dash(self):
    config_file = "aws_dash.yaml"
    print_cloud(config_file=config_file)
    assert True



from verinfast.utils.utils import std_exec


def test_std_exec():
    cmd = ["ls", "-l"]
    result = std_exec(cmd)
    assert result != ""


def test_std_exec_non_unicode():
    cmd = ["printf", b'\x80']
    result = std_exec(cmd)
    assert result == ""



import json
from pathlib import Path
from typing import List

import defusedxml

from verinfast.dependencies.walkers.composer import ComposerWalker
from verinfast.dependencies.walkers.maven import MavenWalker
from verinfast.dependencies.walkers.npm import NodeWalker
from verinfast.dependencies.walkers.gemwalker import GemWalker
from verinfast.dependencies.walkers.nuget import NuGetWalker, c_sharp_matches
from verinfast.dependencies.walkers.dockerwalker import DockerWalker
from verinfast.dependencies.walkers.python import PyWalker
from verinfast.dependencies.walkers.go import GoWalker
from verinfast.dependencies.walkers.classes import Entry

defusedxml.defuse_stdlib()

# Manifests we support
# should probably move this to a conf

# TODO: Gemfile parse
# Example: https://github.com/mastodon/mastodon/blob/main/Gemfile
ruby_matches = ["Gemfile"]

# TODO: Parse Cargo.toml
# Example: https://github.com/servo/servo/blob/master/components/style/Cargo.toml # noqa: E501
rust_matches = ["Cargo.toml"]


def write_file(output_file: str, entries):
    with open(output_file, 'w') as outfile:
        dicts = [entry.to_json() for entry in entries]
        outfile.write(json.dumps(dicts, indent=4))


# Finds all manifests we can process in the repo
# and stores their path in memory
def walk(logger, path: str = "./", output_file="./dependencies.json"):
    mavenWalker = MavenWalker(
        manifest_type="xml",
        manifest_files=["pom.xml"],
        logger=logger,
        root_dir=path
    )

    nodeWalker = NodeWalker(
        manifest_type='json',
        manifest_files=["package.json"],
        logger=logger,
        root_dir=path
    )

    nugetWalker = NuGetWalker(
        manifest_type='xml',
        manifest_files=c_sharp_matches,
        logger=logger,
        root_dir=path
    )

    py_walker = PyWalker(
        manifest_files=["requirements.txt", "requirements-dev.txt"],
        manifest_type="txt",
        logger=logger,
        root_dir=path
    )
    gem_walker = GemWalker(
        manifest_files=["gemfile", "Gemfile"],
        manifest_type="ruby",
        logger=logger,
        root_dir=path
    )
    docker_walker = DockerWalker(
        manifest_files=["Dockerfile", "dockerfile", "docker-compose.yml"],
        manifest_type="Dockerfile",
        logger=logger,
        root_dir=path
    )
    composer_walker = ComposerWalker(
        manifest_files=['composer.json'],
        manifest_type='json',
        logger=logger,
        root_dir=path
    )
    go_walker = GoWalker(
        manifest_type="sum",
        manifest_files=["go.sum"],
        logger=logger,
        root_dir=path
    )
    path = str(Path(path).absolute())
    entries: List[Entry] = []
    composer_walker.initialize(root_path=path)
    entries += composer_walker.entries
    mavenWalker.walk(path=path)
    logger(msg="Dependency Scan 10%", display=True)
    entries += mavenWalker.entries
    write_file(output_file=output_file, entries=entries)
    nodeWalker.initialize(root_path=path)
    logger(msg="Dependency Scan 25%", display=True)
    entries += nodeWalker.entries
    write_file(output_file=output_file, entries=entries)
    nugetWalker.initialize()
    logger(msg="Dependency Scan 40%", display=True)
    nugetWalker.walk(path=path)
    entries += nugetWalker.entries
    write_file(output_file=output_file, entries=entries)
    py_walker.walk(path=path)
    logger(msg="Dependency Scan 60%", display=True)
    entries += py_walker.entries
    write_file(output_file=output_file, entries=entries)
    gem_walker.walk(path=path)
    logger(msg="Dependency Scan 80%", display=True)
    entries += gem_walker.entries
    write_file(output_file=output_file, entries=entries)
    logger(msg="Dependency Scan 95%", display=True)
    docker_walker.walk(path=path, expand=False)
    entries += docker_walker.entries
    write_file(output_file=output_file, entries=entries)
    logger(msg="Dependency Scan 100%", display=True)
    go_walker.walk(path=path)
    entries += go_walker.entries
    write_file(output_file=output_file, entries=entries)
    return output_file



import glob
import json
import subprocess
from pathlib import Path
from typing import List

import httpx


class Entry(dict):
    def __init__(
        self,
        name: str,
        source: str,
        specifier: str = None,
        license: str = None,
        summary: str = None,
        requires=None,
        required_by=None,
    ) -> None:
        if not name:
            raise Exception("Entries must have a package name")
        if not source:
            raise Exception("Entries must have a package source")

        self.name = name
        self.source = source
        self.specifier = specifier
        self.license = license
        self.requires = requires
        self.required_by = required_by
        self.summary = summary

    def __str__(self) -> str:
        d = self.to_json()
        return json.dumps(d, indent=4)

    def __repr__(self) -> str:
        return str(self)

    def to_json(self) -> dict:
        d = {}
        d["name"] = self.name
        d["source"] = self.source
        if self.specifier:
            d["specifier"] = self.specifier
        if self.license:
            d["license"] = self.license
        if self.summary:
            d["summary"] = self.summary
        if self.required_by:
            d["required_by"] = self.required_by
        return d


class Walker():
    def __init__(
        self,
        manifest_type: str,  # "json",
        manifest_files: List[str],  # ["package.json"]
        logger,
        root_dir: str = "./",
        print_name: str = None
    ) -> None:
        if print_name:
            logger(print_name)
        self.files = []
        self.manifest_files = manifest_files.copy()
        for f in self.manifest_files:
            if "*" in f:
                expanded = glob.glob(f, root_dir=root_dir)
                self.manifest_files.remove(f)
                self.manifest_files += expanded
        self.manifest_type = manifest_type
        self.entries = []
        self.requestx = httpx.Client(http2=True, timeout=None)
        self.loggerFunc = logger

    def initialize(self, command: str):
        if command:
            subprocess.call(args=command)

    def log(self, msg, tag=None, display=False, timestamp=True):
        self.loggerFunc(
            msg,
            tag=tag,
            display=display,
            timestamp=timestamp
        )

    def getUrl(self, url: str, headers: dict = {}):
        return  (  # noqa: E271
                    self
                        .requestx  # NOQA: E131
                        .get(url=url, headers=headers)
                        .content
                        .decode('utf-8-sig')
                )

    def walk(self,
             path: str = "./",
             parse: bool = True,
             expand: bool = False,
             debug: int = 0):
        for p in Path(path).rglob('**/*'):
            if debug > 1:
                self.log(F"EVALUATING: {p}", display=(debug > 2))
            if p.name in self.manifest_files:
                if debug > 0:
                    self.log(f"FOUND: {p.name}", display=True)
                self.files.append(p)
                if parse:
                    self.parse(file=str(p.absolute()), expand=expand)

    def parse(self, file: str, expand=False):
        raise Exception("No parser for this Walker")

    def expand(self, file):
        raise Exception("No expansion for this Walker")
